import tempfile
import unittest
from pathlib import Path
import pandas as pd
import pyterrier_caching

class TestIndexerCache(unittest.TestCase):
    def test_basic(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            cache = pyterrier_caching.IndexerCache(d/'cache')
            with self.subTest('len before built'):
                with self.assertRaises(RuntimeError):
                    len(cache)
            cache.index([
                {'docno': '1', 'data': 'test'},
                {'docno': '2', 'data': 'caching pyterrier'},
                {'docno': '3', 'data': 'fetch me some data'},
                {'docno': '4', 'data': 'information retrieval'},
                {'docno': '5', 'data': 'foo bar baz'},
            ])
            with self.subTest('len when built'):
                self.assertEqual(len(cache), 5)
            with self.subTest('full iter'):
                self.assertEqual(list(cache), [
                    {'docno': '1', 'data': 'test'},
                    {'docno': '2', 'data': 'caching pyterrier'},
                    {'docno': '3', 'data': 'fetch me some data'},
                    {'docno': '4', 'data': 'information retrieval'},
                    {'docno': '5', 'data': 'foo bar baz'},
                ])
            with self.subTest('start only'):
                self.assertEqual(list(cache.get_corpus_iter(start=2)), [
                    {'docno': '3', 'data': 'fetch me some data'},
                    {'docno': '4', 'data': 'information retrieval'},
                    {'docno': '5', 'data': 'foo bar baz'},
                ])
            with self.subTest('stop only'):
                self.assertEqual(list(cache.get_corpus_iter(stop=2)), [
                    {'docno': '1', 'data': 'test'},
                    {'docno': '2', 'data': 'caching pyterrier'},
                ])
            with self.subTest('start and stop'):
                self.assertEqual(list(cache.get_corpus_iter(start=1, stop=3)), [
                    {'docno': '2', 'data': 'caching pyterrier'},
                    {'docno': '3', 'data': 'fetch me some data'},
                ])
            with self.subTest('empty'):
                self.assertEqual(list(cache.get_corpus_iter(start=3, stop=1)), [])
            with self.subTest('text_loader'):
                loader = cache.text_loader()
                pd.testing.assert_frame_equal(loader(pd.DataFrame([{'docno': '1'}])), pd.DataFrame([
                    {'docno': '1', 'data': 'test'},
                ]))
                pd.testing.assert_frame_equal(loader(pd.DataFrame([{'docno': '2'}])), pd.DataFrame([
                    {'docno': '2', 'data': 'caching pyterrier'},
                ]))
                pd.testing.assert_frame_equal(loader(pd.DataFrame([{'docno': '2'}, {'docno': '5'}, {'docno': '2'}])), pd.DataFrame([
                    {'docno': '2', 'data': 'caching pyterrier'},
                    {'docno': '5', 'data': 'foo bar baz'},
                    {'docno': '2', 'data': 'caching pyterrier'},
                ]))
                self.assertEqual(len(loader(pd.DataFrame([], columns=['docno']))), 0)

    def test_temp(self):
        with pyterrier_caching.IndexerCache() as cache:
            with self.subTest('len before built'):
                with self.assertRaises(RuntimeError):
                    len(cache)
            cache.index([
                {'docno': '1', 'data': 'test'},
                {'docno': '2', 'data': 'caching pyterrier'},
                {'docno': '3', 'data': 'fetch me some data'},
                {'docno': '4', 'data': 'information retrieval'},
                {'docno': '5', 'data': 'foo bar baz'},
            ])
            with self.subTest('len when built'):
                self.assertEqual(len(cache), 5)
            with self.subTest('full iter'):
                self.assertEqual(list(cache), [
                    {'docno': '1', 'data': 'test'},
                    {'docno': '2', 'data': 'caching pyterrier'},
                    {'docno': '3', 'data': 'fetch me some data'},
                    {'docno': '4', 'data': 'information retrieval'},
                    {'docno': '5', 'data': 'foo bar baz'},
                ])
            with self.subTest('start only'):
                self.assertEqual(list(cache.get_corpus_iter(start=2)), [
                    {'docno': '3', 'data': 'fetch me some data'},
                    {'docno': '4', 'data': 'information retrieval'},
                    {'docno': '5', 'data': 'foo bar baz'},
                ])
            with self.subTest('stop only'):
                self.assertEqual(list(cache.get_corpus_iter(stop=2)), [
                    {'docno': '1', 'data': 'test'},
                    {'docno': '2', 'data': 'caching pyterrier'},
                ])
            with self.subTest('start and stop'):
                self.assertEqual(list(cache.get_corpus_iter(start=1, stop=3)), [
                    {'docno': '2', 'data': 'caching pyterrier'},
                    {'docno': '3', 'data': 'fetch me some data'},
                ])
            with self.subTest('empty'):
                self.assertEqual(list(cache.get_corpus_iter(start=3, stop=1)), [])
            with self.subTest('text_loader'):
                loader = cache.text_loader()
                pd.testing.assert_frame_equal(loader(pd.DataFrame([{'docno': '1'}])), pd.DataFrame([
                    {'docno': '1', 'data': 'test'},
                ]))
                pd.testing.assert_frame_equal(loader(pd.DataFrame([{'docno': '2'}])), pd.DataFrame([
                    {'docno': '2', 'data': 'caching pyterrier'},
                ]))
                pd.testing.assert_frame_equal(loader(pd.DataFrame([{'docno': '2'}, {'docno': '5'}, {'docno': '2'}])), pd.DataFrame([
                    {'docno': '2', 'data': 'caching pyterrier'},
                    {'docno': '5', 'data': 'foo bar baz'},
                    {'docno': '2', 'data': 'caching pyterrier'},
                ]))
                self.assertEqual(len(loader(pd.DataFrame([], columns=['docno']))), 0)
        self.assertFalse(cache.built())

    def test_empty(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            cache = pyterrier_caching.IndexerCache(d/'cache')
            cache.index(range(0))
            with self.subTest('len'):
                self.assertEqual(len(cache), 0)
            with self.subTest('iter'):
                self.assertEqual(list(cache), [])



import pandas as pd
import tempfile
import unittest
from pathlib import Path
import pyterrier as pt
import pyterrier_caching


def raises(df):
    raise RuntimeError()


class TestSparseScorerCache(unittest.TestCase):
    def test_basic(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            scorer_1 = pt.apply.score(lambda df: 1.)
            scorer_raises = pt.apply.score(raises)
            scorer_2 = pt.apply.score(lambda df: 2.)
            cache = pyterrier_caching.KeyValueCache(d/'cache', scorer_1, key=['query', 'docno'], value=['score'])
            # seeding the cache:
            res = cache(pd.DataFrame([
                {'qid': 'a', 'query': 'a', 'docno': '1'},
                {'qid': 'a', 'query': 'a', 'docno': '2'},
                {'qid': 'b', 'query': 'b', 'docno': '2'},
            ]))
            pd.testing.assert_frame_equal(res, pd.DataFrame([
                {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1.},
                {'qid': 'a', 'query': 'a', 'docno': '2', 'score': 1.},
                {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1.},
            ]))

            with self.subTest('make sure it doesn\'t call the transformer again on any of these'):
                cache.transformer = scorer_raises
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '2'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1.},
                    {'qid': 'a', 'query': 'a', 'docno': '2', 'score': 1.},
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1.},
                ]))

            with self.subTest('make sure it calls a new transformer on any unknown values'):
                cache.transformer = scorer_2
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '3'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                    {'qid': 'b', 'query': 'a', 'docno': '1'}, # should pull from (a,1) cache (not dependent on qid)
                    {'qid': 'c', 'query': 'c', 'docno': '1'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1.}, # old
                    {'qid': 'a', 'query': 'a', 'docno': '3', 'score': 2.}, # new
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1.}, # old
                    {'qid': 'b', 'query': 'a', 'docno': '1', 'score': 1.}, # old
                    {'qid': 'c', 'query': 'c', 'docno': '1', 'score': 2.}, # new
                ]))

            cache = None
            with self.subTest('reload the cache and make sure we get the same results'):
                cache = pyterrier_caching.KeyValueCache(d/'cache', scorer_2)
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '3'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                    {'qid': 'b', 'query': 'a', 'docno': '1'}, # should pull from (a,1) cache (not dependent on qid)
                    {'qid': 'c', 'query': 'c', 'docno': '1'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1.}, # old
                    {'qid': 'a', 'query': 'a', 'docno': '3', 'score': 2.}, # new
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1.}, # old
                    {'qid': 'b', 'query': 'a', 'docno': '1', 'score': 1.}, # old
                    {'qid': 'c', 'query': 'c', 'docno': '1', 'score': 2.}, # new
                ]))

            with self.subTest('no scorer necessary if everything is already in the cache'):
                cache = pyterrier_caching.KeyValueCache(d/'cache')
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '3'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                    {'qid': 'b', 'query': 'a', 'docno': '1'}, # should pull from (a,1) cache (not dependent on qid)
                    {'qid': 'c', 'query': 'c', 'docno': '1'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1.}, # old
                    {'qid': 'a', 'query': 'a', 'docno': '3', 'score': 2.}, # new
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1.}, # old
                    {'qid': 'b', 'query': 'a', 'docno': '1', 'score': 1.}, # old
                    {'qid': 'c', 'query': 'c', 'docno': '1', 'score': 2.}, # new
                ]))

            with self.subTest('raises error if document is requested that wasn\'t in the cache and no scorer is provided'):
                with self.assertRaises(LookupError):
                    cache(pd.DataFrame([
                        {'qid': 'a', 'query': 'a', 'docno': '4'},
                        {'qid': 'a', 'query': 'a', 'docno': '3'},
                        {'qid': 'b', 'query': 'b', 'docno': '2'},
                        {'qid': 'b', 'query': 'a', 'docno': '1'},
                        {'qid': 'c', 'query': 'c', 'docno': '1'},
                    ]))



from pathlib import Path
from contextlib import contextmanager
import numpy as np
import pandas as pd
import pyterrier as pt

@contextmanager
def closing_memmap(*args, **kwargs):
    """A context manager that creates a :class:`numpy.memmap` and closes it when the context is exited.

    This allows :class:`numpy.memmap` to be used as a context manager, since it doesn't support the
    context manager protocol directly.

    Args:
        *args: Positional arguments to pass to :class:`numpy.memmap`.
        **kwargs: Keyword arguments to pass to :class:`numpy.memmap`.
    
    Example:

    .. code-block:: python
        :caption: Using a :func:`~pyterrier_caching.util.closing_memmap` context manager.

        from pyterrier_caching import closing_memmap
        with closing_memmap('file.npy', dtype='float32', mode='w+', shape=(10, 10)) as mmp:
            # do what you want with mmp here!
        # mmp is closed here
    """
    mmp = None
    try:
        mmp = np.memmap(*args, **kwargs)
        yield mmp
    finally:
        if mmp is not None:
            del mmp
            mmp = None


class Lazy(pt.Transformer):
    """A :class:`~pyterrier.Transformer` that doesn't initialize until it is used.

    This is useful in cases where loading a transformer is lengthy or allocates resources that are not always
    necessary. For instance a cached neural neural scorer allocates GPU memory, but often isn't needed when used
    with a :class:`~pyterrier_caching.ScorerCache`.

    Example:

    .. code-block:: python
        :caption: Using a :class:`~pyterrier_caching.Lazy` :class:`~pyterrier_dr.ElectraScorer` with a :class:`~pyterrier_caching.ScorerCache`.

        from pyterrier_caching import ScorerCache
        from pyterrier_dr import ElectraScorer
        lazy_scorer = Lazy(ElectraScorer) # ElectraScorer not loaded yet
        cached_scorer = ScorerCache('electra.cache', lazy_scorer)
        cached_scorer([{
            'qid': '0',
            'query': 'terrier breeds',
            'docno': 'doc1',
            'text': 'There are many breeds of terriers, including the Scottish and Jack Russell Terrier.'
        ])
        # ElectraScorer only loaded if ('0', 'doc1') is not yet in electra.cache
    """
    def __init__(self, fn_transformer: pt.Transformer, *fn_args, **fn_kwargs):
        """
        Args:
            fn_transformer: A function that returns a transformer when called (or the transformer class itself).
            fn_args: Positional arguments to pass to ``fn_transformer`` when loading it.
            fn_kwargs: Keyword arguments to pass to ``fn_transformer`` when loading it.
        """
        self.fn_transformer = fn_transformer
        self.fn_args = fn_args
        self.fn_kwargs = fn_kwargs
        self._transformer = None

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        return self.load()(inp)

    def load(self) -> pt.Transformer:
        """Load the transformer if it isn't already loaded, and return it."""
        if not self.loaded():
            self._transformer = self.fn_transformer(*self.fn_args, **self.fn_kwargs)
        return self._transformer

    def unload(self):
        """Unloads the transformer. Subsequent calls to :meth:`load` will re-load it."""
        self._transformer = None

    def loaded(self) -> bool:
        """Return whether the transformer is currently loaded."""
        return self._transformer is not None


def meta_file_compat(path):
    """
    Until version 0.1.0, pt_meta.json was called meta.json. To ensure compatiblity between caches created with
    version <0.1.0 and >=0.1.0, this method moves meta.json to pt_meta.json and linkns meta.json -> pt_meta.json.

    The end effect is that caches created with version <0.1.0 will be compatible with >=0.1.0, but caches created
    with >=0.1.0 will NOT be compatible with those created with <0.1.0.
    """
    path = Path(path)
    if (old_path := (path/'meta.json')).exists() and \
       not (new_path := (path/'pt_meta.json')).exists():
        old_path.rename(new_path)
        old_path.symlink_to(new_path)



from typing import Optional, Union
from pathlib import Path
from tempfile import TemporaryDirectory
from warnings import warn
import hashlib
import lz4.frame
import pandas as pd
import pyterrier as pt
import pickle
import json
import dbm.dumb
import pyterrier_alpha as pta
from pyterrier_caching import meta_file_compat


class DbmRetrieverCache(pta.Artifact, pt.Transformer):
    """A :class:`~pyterrier_caching.RetrieverCache` that stores retrieved results in ``dbm.dumb`` database files."""
    ARTIFACT_TYPE = 'retriever_cache'
    ARTIFACT_FORMAT = 'dbm.dumb'

    def __init__(self,
                 path: Optional[Union[str, Path]] = None,
                 retriever: Optional[pt.Transformer] = None,
                 *,
                 on: Optional[str] = None,
                 verbose: bool = False):
        """
        Args:
            path: The path to the cache.
            retriever: The retriever that is cached.
            on: The column(s) to use as the key for the cache. If None, all columns will be used.
            verbose: If True, print progress information.
        """
        if path is None:
            self._tmpdir = TemporaryDirectory()
            path = Path(self._tmpdir.name) / 'cache'
        else:
            self._tmpdir = None
        super().__init__(path)
        meta_file_compat(path)
        self.on = on
        self.retriever = retriever
        self.verbose = verbose
        self.meta = None
        self.file = None
        self.file_name = None
        if not (Path(self.path)/'pt_meta.json').exists():
            with pta.ArtifactBuilder(self):
                pass # just create the artifact

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        if self.on is not None:
            if isinstance(self.on, str):
                on = [self.on]
            else:
                on = list(self.on)
            pta.validate.columns(inp, includes=on)
        else:
            on = list(inp.columns)
            pta.validate.query_frame(inp, warn=True)
        on = tuple(sorted(on))

        self._ensure_built(on)
        results = []
        to_retrieve = []
        to_retrieve_hashes = []

        # Step 1: Check Cache
        for i in range(len(inp)):
            row = inp.iloc[i]
            key = tuple(row[o] for o in on)
            key_hash = hashlib.sha256(pickle.dumps(key)).digest()
            if key_hash in self.file:
                stored_data = pickle.loads(lz4.frame.decompress(self.file[key_hash]))
                results.append(pd.DataFrame(stored_data))
            else:
                to_retrieve.append(i)
                to_retrieve_hashes.append(key_hash)

        # Step 2: Retrieve and save missing results
        if to_retrieve:
            self.file.close()
            self.file = None
            with dbm.dumb.open(self.file_name, 'w') as file:
                self.file_name = None
                out_cols = pta.inspect.transformer_outputs(self.retriever, list(inp.columns), strict=False)
                if out_cols is not None and all(o in out_cols for o in on):
                    one_at_a_time = False
                    retrieve_phases = [to_retrieve]
                else:
                    warn("Running retriever one query at a time because retriever's outputs could not be determined or "
                         f"the outputs do not contain the cache key: {on}")
                    one_at_a_time = True
                    retrieve_phases = [[idx] for idx in to_retrieve]
                    if self.verbose:
                        retrieve_phases = pt.tqdm(retrieve_phases, unit='q', desc=f'{self}')
                for i, idxs in enumerate(retrieve_phases):
                    retrieved_results = self.retriever(inp.iloc[idxs])
                    retrieved_results.reset_index(drop=True, inplace=True)
                    results.append(retrieved_results)
                    if one_at_a_time:
                        hash_groups = [(to_retrieve_hashes[i], retrieved_results)]
                    else:
                        keys = retrieved_results[list(on)].itertuples(index=False)
                        key_hashes = [hashlib.sha256(pickle.dumps(tuple(key))).digest() for key in keys]
                        hash_groups = retrieved_results.groupby(key_hashes)
                    for key_hash, group in hash_groups:
                        if isinstance(key_hash, tuple):
                            assert len(key_hash) == 1
                            key_hash = key_hash[0]
                        stored_data = {c: group[c].values for c in group.columns}
                        file[key_hash] = lz4.frame.compress(pickle.dumps(stored_data))

        if self.verbose:
            print(f'{self}: {len(inp)-len(to_retrieve)} hit(s), {len(to_retrieve)} miss(es)')
        return pd.concat(results, ignore_index=True)

    def _ensure_built(self, on):
        on_hash = hashlib.sha256(pickle.dumps(on)).hexdigest()
        fname = str(self.path/f'{on_hash}.dbm')
        if self.file_name is not None and self.file_name != fname:
            self.file.close()
            self.file = None
            self.file_name = None
        if self.file is None:
            self.file = dbm.dumb.open(fname, 'c')
            self.file_name = fname
        if self.meta is None:
            with (self.path/'pt_meta.json').open('rt') as fin:
                self.meta = json.load(fin)
        assert self.meta['type'] == self.ARTIFACT_TYPE
        assert self.meta['format'] == self.ARTIFACT_FORMAT

    def close(self):
        if self.file is not None:
            self.file.close()
            self.file = None
            self.file_name = None
        if self._tmpdir is not None:
            self._tmpdir.cleanup()
            self._tmpdir = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __repr__(self):
        return f'DbmRetrieverCache({repr(str(self.path))}, {self.retriever})'


# Default implementation of RetrieverCache: DbmRetrieverCache
RetrieverCache = DbmRetrieverCache



from typing import Optional, Union
from pathlib import Path
from tempfile import TemporaryDirectory
import hashlib
import pickle
import numpy as np
import pandas as pd
import pyterrier as pt
import shutil
import json
import sqlite3
from collections import defaultdict
from contextlib import closing, contextmanager
from more_itertools import chunked
from npids import Lookup
from pyterrier_caching import meta_file_compat
import pyterrier_alpha as pta


class Hdf5ScorerCache(pta.Artifact, pt.Transformer):
    """A cache for storing and retrieving scores for documents, backed by an HDF5 file. 

    This is a *dense* scorer cache, meaning that space for all documents is allocated ahead of time.
    Dense caches are more suitable than sparse ones when a large proportion of the corpus (or
    the entire corpus) is expected to be scored. If only a small proportion of the corpus is expected
    to be scored, a sparse cache (e.g., :class:`~pyterrier_caching.Sqlite3ScorerCache`) may be more appropriate.
    """
    ARTIFACT_TYPE = 'scorer_cache'
    ARTIFACT_FORMAT = 'hdf5'

    def __init__(self,
        path: Optional[Union[str, Path]] = None,
        scorer: Optional[pt.Transformer] = None,
        *,
        verbose: bool = False,
    ):
        """ Creates a new Hdf5ScorerCache instance.

        Args:
            path: The path to the directory where the cache should be stored, or None to create a temporary cache.
            scorer: The scorer to use to score documents that are missing from the cache.
            verbose: Whether to print verbose output.
        """
        if path is None:
            self._tmpdir = TemporaryDirectory()
            path = Path(self._tmpdir.name) / 'cache'
        else:
            self._tmpdir = None
        super().__init__(path)
        meta_file_compat(path)
        self.scorer = scorer
        self.verbose = verbose
        self.meta = None
        self.file = None
        self.docnos = None
        self.dataset_cache = {}

    def transform(self, inp):
        return self.cached_scorer()(inp)

    def built(self) -> bool:
        """Returns whether this cache has been built."""
        return (Path(self.path)/'pt_meta.json').exists()

    def build(self, corpus_iter=None, docnos_file=None):
        """Builds this cache."""
        assert not self.built(), "this cache is alrady built"
        assert corpus_iter is not None or docnos_file is not None
        import h5py
        with pta.ArtifactBuilder(self) as builder:
            with h5py.File(str(builder.path/'data.h5'), 'a'):
                pass # just create the data file
            if docnos_file:
                shutil.copy(docnos_file, builder.path/'docnos.npids')
                builder.metadata['doc_count'] = len(Lookup(builder.path/'docnos.npids'))
            else:
                builder.metadata['doc_count'] = 0
                with Lookup.builder(builder.path/'docnos.npids') as docno_lookup:
                    for record in corpus_iter:
                        docno_lookup.add(record['docno'])
                        builder.metadata['doc_count'] += 1

    def _ensure_built(self):
        import h5py
        assert self.built(), "you must .build(...) this cache before it can be used"
        if self.file is None:
            self.file = h5py.File(self.path/'data.h5', 'r')
        if self.meta is None:
            with (self.path/'pt_meta.json').open('rt') as fin:
                self.meta = json.load(fin)
        if self.docnos is None:
            self.docnos = Lookup(self.path/'docnos.npids')

    def _get_dataset(self, qid):
        if qid not in self.dataset_cache:
            if qid not in self.file:
                return None
            self.dataset_cache[qid] = self.file[qid][:]
        return self.dataset_cache[qid]

    def corpus_count(self) -> int:
        """Returns the number of documents in the corpus that this cache was built from."""
        self._ensure_built()
        return self.meta['doc_count']

    def __repr__(self):
        return f'Hdf5ScorerCache({repr(str(self.path))}, {self.scorer})'

    def cached_scorer(self) -> pt.Transformer:
        """Returns a scorer that uses this cache to store and retrieve scores."""
        return Hdf5ScorerCacheScorer(self)

    def cached_retriever(self, num_results: int = 1000) -> pt.Transformer:
        """Returns a retriever that uses this cache to store and retrieve scores for every document in the corpus.

        This transformer will raie an error if the entire corpus is not scored (e.g., from :meth:`score_all`).
        """
        return Hdf5ScorerCacheRetriever(self, num_results)

    def close(self, delete_tmp: bool = True):
        """ Closes this cache, releasing the file pointer that it holds and writing any new results to disk. """
        if self.file is not None:
            self.dataset_cache = {} # reset the dataset cache
            self.file.close()
            self.file = None
        if self.meta is not None:
            self.meta = None
        if self._tmpdir is not None and delete_tmp:
            self._tmpdir.cleanup()
            self._tmpdir = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def merge_from(self, other: 'Hdf5ScorerCache'):
        """ Merges the cached values from another Hdf5ScorerCache instance into this one.

        Any records that appear in both ``self`` and ``other`` will be replaced with the value from ``other``.
        """
        count = 0
        other._ensure_built()
        self._ensure_built()
        self._ensure_write_mode()
        assert self.meta['doc_count'] == other.meta['doc_count'], f'{self} and {other} are incompatible'
        for key in other.file.keys():
            data = other[key][:]
            if key in self.file:
                # copy over the non-missing values
                mask = ~np.isnan(data)
                self.file[key][mask] = data[mask]
                if self.verbose:
                    count += mask.sum()
            else:
                # easy: just copy it all over
                self.file.create_dataset(key, data=data)
                if self.verbose:
                    count += (~np.isnan(data)).sum()
        if self.verbose:
            print(f"merged {count} records from {other} into {self}")

    def score_all(self, dataset, *, batch_size: int = 1024):
        """Scores all topics for the entire corpus, storing the results in this cache."""
        if not self.built():
            self.build(dataset.get_corpus_iter())
        topics = dataset.get_topics()
        scorer = self.cached_scorer()
        for doc_batch in chunked(dataset.get_corpus_iter(), batch_size):
            doc_batch = pd.DataFrame(doc_batch)
            inp = pd.merge(topics, doc_batch, how='cross')
            scorer(inp)


class Hdf5ScorerCacheScorer(pt.Transformer):
    def __init__(self, cache: Hdf5ScorerCache):
        self.cache = cache

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        import h5py
        self.cache._ensure_built()

        to_score_idxs = []
        to_score_map = defaultdict(list)

        inp = inp.reset_index(drop=True)
        values = pd.Series(index=inp.index, dtype=float)

        # First pass: load what we can from cache
        for query, group in inp.groupby('query'):
            query_hash = hashlib.sha256(query.encode()).hexdigest()
            ds = self.cache._get_dataset(query_hash)
            if ds is None:
                for idx in group.index:
                    to_score_idxs.append(idx)
                    to_score_map[query_hash, group.docno[idx]].append(idx)
            else:
                dids = self.cache.docnos.inv[np.array(group.docno)]
                assert not (dids == -1).any(), 'unknown docno encountered'
                scores = ds[dids]
                for idx, score, is_miss in zip(group.index, scores, np.isnan(scores)):
                    if is_miss:
                        to_score_idxs.append(idx)
                        to_score_map[query_hash, group.docno[idx]].append(idx)
                    else:
                        values[idx] = score

        # Second pass: score the missing ones and add to cache
        if to_score_idxs:
            if self.cache.scorer is None:
                raise LookupError('values missing from cache, but no scorer provided')
            scored = self.cache.scorer(inp.loc[to_score_idxs])
            self.cache.close(delete_tmp=False)
            with h5py.File(self.cache.path/'data.h5', 'a') as fout:
                records = scored[['query', 'docno', 'score']]
                for query, group in records.groupby('query'):
                    query_hash = hashlib.sha256(query.encode()).hexdigest()
                    if query_hash not in fout:
                        fout.create_dataset(query_hash, shape=(self.cache.corpus_count(),), dtype=np.float32, fillvalue=float('nan'))
                    dids = self.cache.docnos.inv[np.array(group.docno)]
                    assert not (dids == -1).any(), 'unknown docno encountered'
                    dids_sorted, dids_sorted_idx = np.unique(dids, return_index=True)
                    fout[query_hash][dids_sorted] = group.score.iloc[dids_sorted_idx]
                    for _, docno, score in group.itertuples(index=False):
                        for idx in to_score_map[query_hash, docno]:
                            values[idx] = score

        results = inp.assign(score=values)
        pt.model.add_ranks(results)
        if self.cache.verbose:
            print(f"{self}: {len(inp)-len(to_score_idxs)} hit(s), {len(to_score_idxs)} miss(es)")
        return results


class Hdf5ScorerCacheRetriever(pt.Transformer):
    def __init__(self, cache: Hdf5ScorerCache, num_results: int = 1000):
        self.cache = cache
        self.num_results = num_results

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        self.cache._ensure_built()
        pta.validate.query_frame(inp, extra_columns=['query'])
        inp = inp.reset_index(drop=True)
        builder = pta.DataFrameBuilder(['_index', 'docno', 'score', 'rank'])
        for i, query in enumerate(inp['query']):
            query_hash = hashlib.sha256(query.encode()).hexdigest()
            ds = self.cache._get_dataset(query_hash)
            if ds is None:
                raise RuntimeError(f'retriever only works if corpus is scored completely; '
                                   f'no cached documents found for query {query!r}.')
            ds = ds[:]
            nans = np.isnan(ds)
            if nans.any():
                raise RuntimeError(f'retriever only works if corpus is scored completely; '
                                   f'{nans.sum()} uncached documents found for query {query!r}.')
            k = min(len(ds), self.num_results)
            docids = ds.argpartition(-k)[-k:]
            scores = ds[docids]
            idxs = scores.argsort()[::-1]
            builder.extend({
                '_index': i,
                'docno': self.cache.docnos.fwd[docids[idxs]],
                'score': scores[idxs].astype(np.float64),
                'rank': np.arange(scores.shape[0]),
            })
        return builder.to_df(merge_on_index=inp)


class Sqlite3ScorerCache(pta.Artifact, pt.Transformer):
    """A cache for storing and retrieving scores for documents, backed by a SQLite3 database.

    This is a *sparse* scorer cache, meaning that space is only allocated for documents that have been scored.
    If a large proportion of the corpus is expected to be scored, a dense cache (e.g., :class:`~pyterrier_caching.Hdf5ScorerCache`)
    may be more appropriate.
    """

    ARTIFACT_TYPE = 'scorer_cache'
    ARTIFACT_FORMAT = 'sqlite3'

    def __init__(
        self,
        path: Optional[Union[str, Path]] = None,
        scorer: pt.Transformer = None,
        *,
        group: Optional[str] = None,
        key: Optional[str] = None,
        value: Optional[str] = None,
        pickle : Optional[bool] = None,
        verbose: bool = False,
    ):
        """
        Args:
            path: The path to the directory where the cache should be stored, or None to create a temporary cache.
            scorer: The scorer to use to score documents that are missing from the cache.
            group: The name of the column in the input DataFrame that contains the group identifier (default: ``query``)
            key: The name of the column in the input DataFrame that contains the document identifier (default: ``docno``)
            value: The name of the column in the input DataFrame that contains the value to cache (default: ``score``)
            pickle: Whether to pickle the value before storing it in the cache (default: False)
            verbose: Whether to print verbose output when scoring documents.

        If a cache does not yet exist at the provided ``path``, a new one is created.

        .. versionchanged:: 0.3.0 added ``pickle`` option to support caching non-numeric values
        """
        if path is None:
            self._tmpdir = TemporaryDirectory()
            path = Path(self._tmpdir.name) / 'cache'
        else:
            self._tmpdir = None
        super().__init__(path)
        meta_file_compat(path)
        self.scorer = scorer
        self.verbose = verbose
        self.meta = None
        if not (Path(self.path)/'pt_meta.json').exists():
            if group is None:
                group = 'query'
            if key is None:
                key = 'docno'
            if value is None:
                value = 'score'
            with pta.ArtifactBuilder(self) as builder:
                builder.metadata['group'] = group
                builder.metadata['key'] = key
                builder.metadata['value'] = value
                builder.metadata['pickle'] = pickle
                self.db = sqlite3.connect(builder.path/'db.sqlite3')
                value_type = "BLOB" if pickle else "NUMERIC"
                with closing(self.db.cursor()) as cursor:
                    cursor.execute(f"""
                        CREATE TABLE IF NOT EXISTS cache (
                          [group] TEXT NOT NULL,
                          key TEXT NOT NULL,
                          value {value_type} NOT NULL,
                          PRIMARY KEY ([group], key)
                        )
                    """)
        else:
            self.db = sqlite3.connect(self.path/'db.sqlite3')
        with (Path(self.path)/'pt_meta.json').open('rt') as fin:
            self.meta = json.load(fin)
        self.meta.setdefault('pickle', False)
        if group is not None:
            assert group == self.meta['group'], f'group={group!r} provided, but index created with group={self.meta["group"]!r}'
        self.group = self.meta['group']
        if key is not None:
            assert key == self.meta['key'], f'key={key!r} provided, but index created with key={self.meta["key"]!r}'
        self.key = self.meta['key']
        if value is not None:
            assert value == self.meta['value'], f'value={value!r} provided, but index created with value={self.meta["value"]!r}'
        self.value = self.meta['value']
        if pickle is not None:
            assert pickle == self.meta['pickle'], f'pickle={pickle!r} provided, but index created with pickle={self.meta["pickle"]!r}'
        self.pickle = self.meta['pickle']

    def close(self):
        """ Closes this cache, releasing the sqlite connection that it holds. """
        if self.db is not None:
            self.db.close()
            self.db = None
        if self._tmpdir is not None:
            self._tmpdir.cleanup()
            self._tmpdir = None

    @contextmanager
    def _cursor(self):
        assert self.db is not None, "cache is closed"
        with closing(self.db.cursor()) as cursor:
            yield cursor

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        """ Scores the input DataFrame using cached values, scoring any missing ones and adding them to the cache. """
        pta.validate.columns(inp, includes=[self.group, self.key])

        to_score_idxs = []
        to_score_map = {}

        inp = inp.reset_index(drop=True)
        values = pd.Series(index=inp.index, dtype=object if self.pickle else float)

        # First pass: load what we can from cache
        for group_key, group in inp.groupby(self.group):
            placeholder = ', '.join(['?'] * len(group))
            key2idxs = defaultdict(list)
            for idx, key in zip(group.index, group[self.key]):
                key2idxs[key].append(idx)
            with self._cursor() as cursor:
                cursor.execute(f'SELECT key, value FROM cache WHERE [group]=? AND key IN ({placeholder})',
                    [group_key] + group[self.key].tolist())
                for key, score in cursor.fetchall():
                    for idx in key2idxs[key]:
                        values[idx] = pickle.loads(score) if self.pickle else score 
                    del key2idxs[key]
            for key, idxs in key2idxs.items():
                to_score_idxs.extend(idxs)
                to_score_map[group_key, key] = idxs

        # Second pass: score the missing ones and add to cache
        if to_score_idxs:
            if self.scorer is None:
                raise LookupError('values missing from cache, but no scorer provided')
            scored = self.scorer(inp.loc[to_score_idxs])
            records = scored[[self.group, self.key, self.value]]
            rec_it = records.itertuples(index=False)
            if self.pickle:
                rec_it = [(g, k, pickle.dumps(v)) for g, k, v in rec_it]
            with closing(self.db.cursor()) as cursor:
                cursor.executemany('INSERT INTO cache ([group], key, value) VALUES (?, ?, ?)', rec_it)
                self.db.commit()
            for group, key, score in records.itertuples(index=False):
                for idx in to_score_map[group, key]:
                    values[idx] = score

        results = inp.assign(**{self.value: values})
        if self.value == 'score':
            pt.model.add_ranks(results)
        if self.verbose:
            print(f"{self}: {len(inp)-len(to_score_idxs)} hit(s), {len(to_score_idxs)} miss(es)")
        return results

    def merge_from(self, other: 'Sqlite3ScorerCache'):
        """ Merges the cached values from another Sqlite3ScorerCache instance into this one.

        Any keys that appear in both ``self`` and ``other`` will be replaced with the value from ``other``.
        """
        count = 0
        with self._cursor() as insert_cursor, other._cursor() as select_cursor:
            select_cursor.execute('SELECT [group], key, value FROM cache')
            while batch := select_cursor.fetchmany(10_000):
                count += len(batch)
                insert_cursor.executemany('INSERT OR REPLACE INTO cache ([group], key, value) VALUES (?, ?, ?)', batch)
            self.db.commit()
        if self.verbose:
            print(f"merged {count} records from {other} into {self}")

    def __repr__(self):
        return f'Sqlite3ScorerCache({str(self.path)!r}, {self.scorer!r}, group={self.group!r}, key={self.key!r})'

# Default implementations
ScorerCache = Sqlite3ScorerCache
DenseScorerCache = Hdf5ScorerCache
SparseScorerCache = Sqlite3ScorerCache



from typing import Optional, List, Iterator, Dict, Any, Union, Literal
from pathlib import Path
from contextlib import ExitStack
import pickle
import json
import lz4.frame
from tempfile import TemporaryDirectory
import numpy as np
import pandas as pd
import pyterrier as pt
from npids import Lookup
from tqdm import tqdm
import pyterrier_alpha as pta
from pyterrier_caching import closing_memmap, meta_file_compat


class Lz4PickleIndexerCache(pta.Artifact, pt.Indexer):
    """An :class:`~pyterrier_caching.IndexerCache` that stores records as pickled dictionaries compressed with lz4.
    """
    ARTIFACT_TYPE = 'indexer_cache'
    ARTIFACT_FORMAT = 'lz4pickle'

    def __init__(self, path: Optional[Union[str, Path]] = None):
        """
        Args:
            path: The path to the cache. If None, a temporary cache will be created that is deleted when closed.
        """
        if path is None:
            self._tmpdir = TemporaryDirectory()
            path = Path(self._tmpdir.name) / 'cache'
        else:
            self._tmpdir = None
        super().__init__(path)
        meta_file_compat(path)
        self._docnos = None

    def indexer(self, mode: Union[str, pta.ArtifactBuilderMode] = pta.ArtifactBuilderMode.create, skip_docno_lookup: bool = False) -> pt.Indexer:
        """Returns an :class:`~pyterrier.Indexer` for this cache. The indexer can be used to create the cache.

        Args:
            mode: The mode to use for the indexer. Must be 'create'.
            skip_docno_lookup: If True, skip creating a docno lookup.
        """
        return Lz4PickleIndexerCacheIndexer(self, mode, skip_docno_lookup=skip_docno_lookup)

    def index(self, it: Iterator[Dict[str, Any]]) -> None:
        """Indexes the provided records to this cache."""
        return self.indexer().index(it)

    def __iter__(self) -> Iterator[Dict[str, Any]]:
        """Iterates over the records stored in the cache."""
        return self.get_corpus_iter()

    def __len__(self) -> int:
        """Returns the number of records stored in the cache."""
        if not self.built():
            raise RuntimeError('cache not built')
        with (Path(self.path)/'pt_meta.json').open('rt') as fin:
            metadata = json.load(fin)
        return metadata['record_count']

    def get_corpus_iter(self, verbose: bool = False, fields: Optional[List[str]] = None, start: Optional[int] = None, stop: Optional[int] = None) -> Iterator[Dict[str, Any]]:
        """Iterates over the records stored in the cache.

        Args:
            verbose: If True, show a progress bar.
            fields: If not None, only return these fields.
            start: If not None, start at this record number.
            stop: If not None, stop at this record number.
        """
        # validate arguments
        if fields is not None:
            fields = set(fields)
        if start is not None:
            assert start >= 0
        if stop is not None:
            assert stop >= 0
            # we take one extra one to get the offset of the last record, too.
            stop += 1

        with open(Path(self.path)/'data.pkl.lz4', mode='rb') as fdata, \
             closing_memmap(Path(self.path)/'offsets.np', dtype=np.uint64) as offsets_mmp:
            offsets = offsets_mmp[start:stop]
            if len(offsets) > 0:
                fdata.seek(offsets[0])
                it = zip(offsets[:-1], offsets[1:])
                if verbose:
                    it = tqdm(it, total=len(offsets)-1, unit='d')
                for offset_start, offset_stop in it:
                    record_length = offset_stop - offset_start
                    record = fdata.read(record_length)
                    record = lz4.frame.decompress(record)
                    record = pickle.loads(record)
                    if fields is not None:
                        record = {k: v for k, v in record.items() if k in fields}
                    yield record

    def to_dataframe(self, verbose: bool = False, fields: Optional[List[str]] = None, start: Optional[int] = None, stop: Optional[int] = None) -> pd.DataFrame:
        """Converts the results in this cache to a DataFrame.

        Args:
            verbose: If True, show a progress bar.
            fields: If not None, only return these fields.
            start: If not None, start at this record number.
            stop: If not None, stop at this record number.
        """
        return pd.DataFrame(self.get_corpus_iter(verbose=verbose, fields=fields, start=start, stop=stop))

    def __getitem__(self, items: Union[int, str, slice]):
        """Returns the record(s) stored in the cache by the provided index, docno, or range.
        """
        if isinstance(items, int):
            return next(self.get_corpus_iter(start=items, stop=items+1))
        elif isinstance(items, str):
            idx = Lookup(Path(self.path)/'docnos.npids').inv[items]
            return self[idx]
            # TODO: support lookup by a list of strings?
        elif isinstance(items, slice):
            assert items.step is None, "step is not supported by Lz4PickleIndexerCache"
            return list(self.get_corpus_iter(start=items.start, stop=items.stop))
        raise ValueError('unknown type for items: {}'.format(type(items)))

    def built(self) -> bool:
        """Returns True if the cache is built."""
        return (Path(self.path)/'pt_meta.json').exists()

    def text_loader(
        self,
        fields: Union[str, List[str], Literal['*']] = '*',
        *,
        verbose: bool = False,
    ) -> pt.Transformer:
        """Returns a :class:`~pyterrier.Transformer` that loads the text from the cache based on ``docno``.

        Args:
            fields: If not '*', only return these fields.
            verbose: If True, show a progress bar.
        """
        return Lz4PickleIndexerCacheTextLoader(self, fields, verbose=verbose)

    def docnos(self):
        """Returns a :class:`~npids.Lookup` for the docnos stored in the cache."""
        assert self.built()
        if self._docnos is None:
            self._docnos = Lookup(Path(self.path)/'docnos.npids')
        return self._docnos

    def close(self):
        """Closes any open files used by this cache."""
        if self._docnos is not None:
            self._docnos.close()
            self._docnos = None
        if self._tmpdir is not None:
            self._tmpdir.cleanup()
            self._tmpdir = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class Lz4PickleIndexerCacheIndexer(pt.Indexer):
    def __init__(self, cache: Lz4PickleIndexerCache, mode: Union[str, pta.ArtifactBuilderMode], skip_docno_lookup: bool = False):
        self.cache = cache
        self.mode = mode
        self.skip_docno_lookup = skip_docno_lookup

    def index(self, it: Iterator[Dict[str, Any]]) -> None:
        assert pta.ArtifactBuilderMode(self.mode) == pta.ArtifactBuilderMode.create, "Lz4PickleIndexerCache only supports 'create' mode"
        with ExitStack() as s:
            builder = s.enter_context(pta.ArtifactBuilder(self.cache, mode=self.mode))
            fdata = s.enter_context(open(builder.path/'data.pkl.lz4', mode='wb'))
            foffsets = s.enter_context(open(builder.path/'offsets.np', 'wb'))
            docno_lookup = False if self.skip_docno_lookup else None
            foffsets.write(np.array(0, dtype=np.uint64).tobytes())
            builder.metadata['record_count'] = 0
            for record in it:
                if docno_lookup is None:
                    if 'docno' in record:
                        docno_lookup = s.enter_context(Lookup.builder(builder.path/'docnos.npids'))
                if docno_lookup is not False:
                    docno_lookup.add(record['docno'])
                record_bytes = pickle.dumps(dict(record))
                fdata.write(lz4.frame.compress(record_bytes))
                foffsets.write(np.array(fdata.tell(), dtype=np.uint64))
                builder.metadata['record_count'] += 1


class Lz4PickleIndexerCacheTextLoader(pt.Transformer):
    def __init__(self,
        cache: Lz4PickleIndexerCache,
        fields: Union[str, List[str], Literal['*']] = '*',
        *,
        verbose: bool = False,
    ):
        self.cache = cache
        self.fields = fields
        self.verbose = verbose

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        pta.validate.columns(inp, includes=['docno'])
        if len(inp) == 0:
            return inp

        if self.fields == '*':
            fields = None
        elif isinstance(self.fields, str):
            fields = [self.fields]
        else:
            fields = self.fields

        doc_ids = self.cache.docnos().inv[inp['docno'].to_list()]

        builder = None
        with open(Path(self.cache.path)/'data.pkl.lz4', mode='rb') as fdata, \
             closing_memmap(Path(self.cache.path)/'offsets.np', dtype=np.uint64) as offsets_mmp:
            if self.verbose:
                doc_ids = tqdm(doc_ids, total=len(doc_ids), unit='d')
            for did in doc_ids:
                offset_start, offset_stop = offsets_mmp[did:did+2]
                record_length = offset_stop - offset_start
                fdata.seek(offset_start)
                record = fdata.read(record_length)
                record = lz4.frame.decompress(record)
                record = pickle.loads(record)
                if fields is not None:
                    record = {k: v for k, v in record.items() if k in fields}
                record.pop('docno')
                if builder is None:
                    builder = pta.DataFrameBuilder(list(record.keys()))
                builder.extend(record)

        return builder.to_df(inp)


# Default implementation of IndexerCache: Lz4PickleIndexerCache
IndexerCache = Lz4PickleIndexerCache



# pip install -q git+https://github.com/naver/splade.git git+https://github.com/cmacdonald/pyt_splade.git
from time import time
import pyterrier as pt
pt.init()
from pyterrier_caching import IndexerCache
from pyt_splade import SpladeFactory

cache = IndexerCache('splade.cache')
if not cache.built():
    dataset = pt.get_dataset('irds:antique')
    cache_pipeline = SpladeFactory().indexing() >> cache
    t0 = time()
    cache_pipeline.index(dataset.get_corpus_iter())
    t1 = time()
    print(f'time to cache: {t1-t0:.3f}s')

indexer = pt.IterDictIndexer('./splade.terrier', pretokenised=True)
t0 = time()
indexer.index(iter(cache))
t1 = time()
print(f'time to index from cache: {t1-t0:.3f}s')



from time import time
import pyterrier as pt
pt.init()
from pyterrier_caching import RetrieverCache

# Setup
bm25 = pt.BatchRetrieve.from_dataset('msmarco_passage', 'terrier_stemmed', wmodel='BM25')
cached_retriever = RetrieverCache('bm25.cache', bm25, verbose=True)
dataset = pt.get_dataset('irds:msmarco-passage/dev/small')


t0 = time()
cached_retriever(dataset.get_topics())
t1 = time()
print(f'first invocation: {t1-t0:.3f}s')

t0 = time()
cached_retriever(dataset.get_topics())
t1 = time()
print(f'second invocation: {t1-t0:.3f}s')



import pandas as pd
import numpy as np
import tempfile
import unittest
from pathlib import Path
import pyterrier as pt
import pyterrier_caching

class TestRetrieverCache(unittest.TestCase):
    def test_basic(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            def retriever_1(df):
                res = pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 2., 'rank': 0},
                    {'qid': 'a', 'query': 'a', 'docno': '2', 'score': 1., 'rank': 1},
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 3., 'rank': 0},
                ])
                return res[res['query'].isin(df['query'])]
            def retriever_2(df):
                res = pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '3', 'score': 3., 'rank': 0},
                    {'qid': 'a', 'query': 'a', 'docno': '4', 'score': 2., 'rank': 1},
                    {'qid': 'c', 'query': 'c', 'docno': '0', 'score': 0., 'rank': 0},
                ])
                return res[res['query'].isin(df['query'])]
            retriever_raises = pt.apply.generic(lambda df: 0/0)
            # retriever_2 = pt.apply.generic(lambda df: 2.)
            cache = pyterrier_caching.RetrieverCache(d/'cache', pt.apply.generic(retriever_1))
            # seeding the cache:
            res = cache(pd.DataFrame([
                {'qid': 'a', 'query': 'a'},
                {'qid': 'b', 'query': 'b'},
            ]))
            pd.testing.assert_frame_equal(res, pd.DataFrame([
                {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 2., 'rank': 0},
                {'qid': 'a', 'query': 'a', 'docno': '2', 'score': 1., 'rank': 1},
                {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 3., 'rank': 0},
            ]))

            with self.subTest('make sure it doesn\'t call the retriever again on any of these'):
                cache.retriever = retriever_raises
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a'},
                    {'qid': 'b', 'query': 'b'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 2., 'rank': 0},
                    {'qid': 'a', 'query': 'a', 'docno': '2', 'score': 1., 'rank': 1},
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 3., 'rank': 0},
                ]))

            with self.subTest('only some of the queries'):
                cache.retriever = retriever_raises
                res = cache(pd.DataFrame([
                    {'qid': 'b', 'query': 'b'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 3., 'rank': 0},
                ]))

            with self.subTest('make sure it calls a new scorer on any unknown values'):
                cache.retriever = pt.apply.generic(retriever_2)
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a'},
                    {'qid': 'c', 'query': 'c'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 2., 'rank': 0}, # old
                    {'qid': 'a', 'query': 'a', 'docno': '2', 'score': 1., 'rank': 1}, # old
                    {'qid': 'c', 'query': 'c', 'docno': '0', 'score': 0., 'rank': 0}, # new
                ]))

    def test_keys_not_present_in_output(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            def retriever_1(df):
                res = pd.DataFrame([
                    {'qid': 'a', 'docno': '1', 'score': 2., 'rank': 0},
                    {'qid': 'a', 'docno': '2', 'score': 1., 'rank': 1},
                    {'qid': 'b', 'docno': '2', 'score': 3., 'rank': 0},
                ])
                return res[res['qid'].isin(df['qid'])]
            def retriever_2(df):
                res = pd.DataFrame([
                    {'qid': 'a', 'docno': '3', 'score': 3., 'rank': 0},
                    {'qid': 'a', 'docno': '4', 'score': 2., 'rank': 1},
                    {'qid': 'c', 'docno': '0', 'score': 0., 'rank': 0},
                ])
                return res[res['qid'].isin(df['qid'])]
            retriever_raises = pt.apply.generic(lambda df: 0/0)
            # retriever_2 = pt.apply.generic(lambda df: 2.)
            cache = pyterrier_caching.RetrieverCache(d/'cache', pt.apply.generic(retriever_1))
            # seeding the cache:
            res = cache(pd.DataFrame([
                {'qid': 'a', 'query': 'a'},
                {'qid': 'b', 'query': 'b'},
            ]))
            pd.testing.assert_frame_equal(res, pd.DataFrame([
                {'qid': 'a', 'docno': '1', 'score': 2., 'rank': 0},
                {'qid': 'a', 'docno': '2', 'score': 1., 'rank': 1},
                {'qid': 'b', 'docno': '2', 'score': 3., 'rank': 0},
            ]))

            with self.subTest('make sure it doesn\'t call the retriever again on any of these'):
                cache.retriever = retriever_raises
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a'},
                    {'qid': 'b', 'query': 'b'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'docno': '1', 'score': 2., 'rank': 0},
                    {'qid': 'a', 'docno': '2', 'score': 1., 'rank': 1},
                    {'qid': 'b', 'docno': '2', 'score': 3., 'rank': 0},
                ]))

            with self.subTest('only some of the queries'):
                cache.retriever = retriever_raises
                res = cache(pd.DataFrame([
                    {'qid': 'b', 'query': 'b'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'b', 'docno': '2', 'score': 3., 'rank': 0},
                ]))

            with self.subTest('make sure it calls a new scorer on any unknown values'):
                cache.retriever = pt.apply.generic(retriever_2)
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a'},
                    {'qid': 'c', 'query': 'c'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'docno': '1', 'score': 2., 'rank': 0}, # old
                    {'qid': 'a', 'docno': '2', 'score': 1., 'rank': 1}, # old
                    {'qid': 'c', 'docno': '0', 'score': 0., 'rank': 0}, # new
                ]))



__version__ = '0.5.1'

from pyterrier_caching.util import Lazy, closing_memmap, meta_file_compat
from pyterrier_caching.indexer_cache import IndexerCache, Lz4PickleIndexerCache
from pyterrier_caching.scorer_cache import ScorerCache, DenseScorerCache, SparseScorerCache, Hdf5ScorerCache, Sqlite3ScorerCache
from pyterrier_caching.retriever_cache import RetrieverCache, DbmRetrieverCache
from pyterrier_caching.key_value_cache import KeyValueCache, Sqlite3KeyValueCache

__all__ = ["Lazy", "closing_memmap", "meta_file_compat", "IndexerCache", "Lz4PickleIndexerCache", "ScorerCache", "DenseScorerCache", "SparseScorerCache", "Hdf5ScorerCache", "Sqlite3ScorerCache", "RetrieverCache", "DbmRetrieverCache", 'KeyValueCache', 'Sqlite3KeyValueCache']



import numpy as np
import pandas as pd
import tempfile
import unittest
from pathlib import Path
import pyterrier as pt
import pyterrier_caching
from npids import Lookup


def raises(df):
    raise RuntimeError()


class TestDenseScorerCache(unittest.TestCase):
    def test_build_from_iter(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            cache = pyterrier_caching.DenseScorerCache(d/'cache')
            cache.build([
                {'docno': '1', 'data': 'test'},
                {'docno': '2', 'data': 'caching pyterrier'},
                {'docno': '3', 'data': 'fetch me some data'},
                {'docno': '4', 'data': 'information retrieval'},
                {'docno': '5', 'data': 'foo bar baz'},
            ])
            self.assertEqual(cache.corpus_count(), 5)

    def test_build_from_docnos(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            cache = pyterrier_caching.DenseScorerCache(d/'cache')
            docnos = Lookup.build(['1', '2', '3', '4', '5'], d/'docnos.npids')
            cache.build(docnos_file=d/'docnos.npids')
            self.assertEqual(cache.corpus_count(), 5)

    def test_basic(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            scorer_1 = pt.apply.score(lambda df: 1.)
            scorer_raises = pt.apply.score(raises)
            scorer_2 = pt.apply.score(lambda df: 2.)
            cache = pyterrier_caching.DenseScorerCache(d/'cache', scorer_1)
            docnos = Lookup.build(['1', '2', '3', '4', '5'], d/'docnos.npids')
            cache.build(docnos_file=d/'docnos.npids')
            # seeding the cache:
            res = cache(pd.DataFrame([
                {'qid': 'a', 'query': 'a', 'docno': '1'},
                {'qid': 'a', 'query': 'a', 'docno': '2'},
                {'qid': 'b', 'query': 'b', 'docno': '2'},
            ]))
            pd.testing.assert_frame_equal(res, pd.DataFrame([
                {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 0},
                {'qid': 'a', 'query': 'a', 'docno': '2', 'score': 1., 'rank': 1},
                {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1., 'rank': 0},
            ]))

            with self.subTest('make sure it doesn\'t call the scorer again on any of these'):
                cache.scorer = scorer_raises
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '2'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 0},
                    {'qid': 'a', 'query': 'a', 'docno': '2', 'score': 1., 'rank': 1},
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1., 'rank': 0},
                ]))

            with self.subTest('make sure it calls a new scorer on any unknown values'):
                cache.scorer = scorer_2
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '3'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                    {'qid': 'b', 'query': 'a', 'docno': '1'}, # should pull from (a,1) cache (not dependent on qid)
                    {'qid': 'c', 'query': 'c', 'docno': '1'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 1}, # old
                    {'qid': 'a', 'query': 'a', 'docno': '3', 'score': 2., 'rank': 0}, # new
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1., 'rank': 0}, # old
                    {'qid': 'b', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 1}, # old
                    {'qid': 'c', 'query': 'c', 'docno': '1', 'score': 2., 'rank': 0}, # new
                ]))

            with self.subTest('reload the cache and make sure we get the same results'):
                cache = pyterrier_caching.DenseScorerCache(d/'cache', scorer_2)
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '3'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                    {'qid': 'b', 'query': 'a', 'docno': '1'}, # should pull from (a,1) cache (not dependent on qid)
                    {'qid': 'c', 'query': 'c', 'docno': '1'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 1}, # old
                    {'qid': 'a', 'query': 'a', 'docno': '3', 'score': 2., 'rank': 0}, # new
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1., 'rank': 0}, # old
                    {'qid': 'b', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 1}, # old
                    {'qid': 'c', 'query': 'c', 'docno': '1', 'score': 2., 'rank': 0}, # new
                ]))

            with self.subTest('no scorer necessary if everything is already in the cache'):
                cache = pyterrier_caching.DenseScorerCache(d/'cache')
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '3'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                    {'qid': 'b', 'query': 'a', 'docno': '1'}, # should pull from (a,1) cache (not dependent on qid)
                    {'qid': 'c', 'query': 'c', 'docno': '1'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 1}, # old
                    {'qid': 'a', 'query': 'a', 'docno': '3', 'score': 2., 'rank': 0}, # new
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1., 'rank': 0}, # old
                    {'qid': 'b', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 1}, # old
                    {'qid': 'c', 'query': 'c', 'docno': '1', 'score': 2., 'rank': 0}, # new
                ]))

            with self.subTest('raises error if document is requested that wasn\'t in the cache and no scorer is provided'):
                with self.assertRaises(LookupError):
                    cache(pd.DataFrame([
                        {'qid': 'a', 'query': 'a', 'docno': '4'},
                        {'qid': 'a', 'query': 'a', 'docno': '3'},
                        {'qid': 'b', 'query': 'b', 'docno': '2'},
                        {'qid': 'b', 'query': 'a', 'docno': '1'},
                        {'qid': 'c', 'query': 'c', 'docno': '1'},
                    ]))

    def test_cached_retriever(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            scorer = pt.apply.score(lambda df: float(df['docno']))
            cache = pyterrier_caching.DenseScorerCache(d/'cache', scorer)
            docnos = Lookup.build(['1', '2', '3', '4', '5'], d/'docnos.npids')
            cache.build(docnos_file=d/'docnos.npids')
            # seeding the cache:
            res = cache(pd.DataFrame([
                {'qid': 'a', 'query': 'a', 'docno': '1'},
                {'qid': 'a', 'query': 'a', 'docno': '2'},
                {'qid': 'b', 'query': 'b', 'docno': '2'},
            ]))

            with self.subTest('should raise error when retrieving, since not all docs are cached'):
                with self.assertRaises(RuntimeError):
                    cache.cached_retriever()(pd.DataFrame([
                        {'qid': 'a', 'query': 'a'},
                        {'qid': 'b', 'query': 'b'},
                    ]))

            # completely score query a
            res = cache(pd.DataFrame([
                {'qid': 'a', 'query': 'a', 'docno': '3'},
                {'qid': 'a', 'query': 'a', 'docno': '4'},
                {'qid': 'a', 'query': 'a', 'docno': '5'},
            ]))

            with self.subTest('query a should return cached values'):
                res = cache.cached_retriever(num_results=2)(pd.DataFrame([
                    {'qid': 'a', 'query': 'a'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '5', 'score': 5., 'rank': 0},
                    {'qid': 'a', 'query': 'a', 'docno': '4', 'score': 4., 'rank': 1},
                ]))

            with self.subTest('num_results is robust'):
                res = cache.cached_retriever(num_results=1000)(pd.DataFrame([
                    {'qid': 'a', 'query': 'a'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '5', 'score': 5., 'rank': 0},
                    {'qid': 'a', 'query': 'a', 'docno': '4', 'score': 4., 'rank': 1},
                    {'qid': 'a', 'query': 'a', 'docno': '3', 'score': 3., 'rank': 2},
                    {'qid': 'a', 'query': 'a', 'docno': '2', 'score': 2., 'rank': 3},
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 4},
                ]))

            with self.subTest('query b should still raise an error'):
                with self.assertRaises(RuntimeError):
                    cache.cached_retriever(num_results=2)(pd.DataFrame([
                        {'qid': 'b', 'query': 'b'},
                    ]))

            with self.subTest('should still raise an error when retrieving both'):
                with self.assertRaises(RuntimeError):
                    cache.cached_retriever(num_results=2)(pd.DataFrame([
                        {'qid': 'a', 'query': 'a'},
                        {'qid': 'b', 'query': 'b'},
                    ]))

            with self.subTest('should raise for an unseen query'):
                with self.assertRaises(RuntimeError):
                    cache.cached_retriever(num_results=2)(pd.DataFrame([
                        {'qid': 'c', 'query': 'c'},
                    ]))


class TestSparseScorerCache(unittest.TestCase):
    def test_basic(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            scorer_1 = pt.apply.score(lambda df: 1.)
            scorer_raises = pt.apply.score(raises)
            scorer_2 = pt.apply.score(lambda df: 2.)
            cache = pyterrier_caching.SparseScorerCache(d/'cache', scorer_1)
            # seeding the cache:
            res = cache(pd.DataFrame([
                {'qid': 'a', 'query': 'a', 'docno': '1'},
                {'qid': 'a', 'query': 'a', 'docno': '2'},
                {'qid': 'b', 'query': 'b', 'docno': '2'},
            ]))
            pd.testing.assert_frame_equal(res, pd.DataFrame([
                {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 0},
                {'qid': 'a', 'query': 'a', 'docno': '2', 'score': 1., 'rank': 1},
                {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1., 'rank': 0},
            ]))

            with self.subTest('make sure it doesn\'t call the scorer again on any of these'):
                cache.scorer = scorer_raises
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '2'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 0},
                    {'qid': 'a', 'query': 'a', 'docno': '2', 'score': 1., 'rank': 1},
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1., 'rank': 0},
                ]))

            with self.subTest('make sure it calls a new scorer on any unknown values'):
                cache.scorer = scorer_2
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '3'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                    {'qid': 'b', 'query': 'a', 'docno': '1'}, # should pull from (a,1) cache (not dependent on qid)
                    {'qid': 'c', 'query': 'c', 'docno': '1'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 1}, # old
                    {'qid': 'a', 'query': 'a', 'docno': '3', 'score': 2., 'rank': 0}, # new
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1., 'rank': 0}, # old
                    {'qid': 'b', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 1}, # old
                    {'qid': 'c', 'query': 'c', 'docno': '1', 'score': 2., 'rank': 0}, # new
                ]))

            cache = None
            with self.subTest('reload the cache and make sure we get the same results'):
                cache = pyterrier_caching.SparseScorerCache(d/'cache', scorer_2)
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '3'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                    {'qid': 'b', 'query': 'a', 'docno': '1'}, # should pull from (a,1) cache (not dependent on qid)
                    {'qid': 'c', 'query': 'c', 'docno': '1'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 1}, # old
                    {'qid': 'a', 'query': 'a', 'docno': '3', 'score': 2., 'rank': 0}, # new
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1., 'rank': 0}, # old
                    {'qid': 'b', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 1}, # old
                    {'qid': 'c', 'query': 'c', 'docno': '1', 'score': 2., 'rank': 0}, # new
                ]))

            with self.subTest('no scorer necessary if everything is already in the cache'):
                cache = pyterrier_caching.SparseScorerCache(d/'cache')
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '3'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                    {'qid': 'b', 'query': 'a', 'docno': '1'}, # should pull from (a,1) cache (not dependent on qid)
                    {'qid': 'c', 'query': 'c', 'docno': '1'},
                ]))
                pd.testing.assert_frame_equal(res, pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 1}, # old
                    {'qid': 'a', 'query': 'a', 'docno': '3', 'score': 2., 'rank': 0}, # new
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1., 'rank': 0}, # old
                    {'qid': 'b', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 1}, # old
                    {'qid': 'c', 'query': 'c', 'docno': '1', 'score': 2., 'rank': 0}, # new
                ]))

            with self.subTest('raises error if document is requested that wasn\'t in the cache and no scorer is provided'):
                with self.assertRaises(LookupError):
                    cache(pd.DataFrame([
                        {'qid': 'a', 'query': 'a', 'docno': '4'},
                        {'qid': 'a', 'query': 'a', 'docno': '3'},
                        {'qid': 'b', 'query': 'b', 'docno': '2'},
                        {'qid': 'b', 'query': 'a', 'docno': '1'},
                        {'qid': 'c', 'query': 'c', 'docno': '1'},
                    ]))
    def test_features(self):
        with tempfile.TemporaryDirectory() as d:
            feat_1 = pt.apply.doc_features(lambda row: np.array([0.2, 0.1]) if row["qid"] == "q1" else  np.array([0.4, 0.3]))
            d = Path(d)
            cache = pyterrier_caching.SparseScorerCache(d/'cache', feat_1, value="features", pickle=True)
            
            df1 = pd.DataFrame([{'qid' : 'q1', 'query' : 'a', 'docno' : 'd1'}])
            res = cache.transform(df1)
            res1c = cache.transform(df1)
            self.assertEqual(0.2, res1c.iloc[0]["features"][0])
            pd.testing.assert_frame_equal(res, res1c)

            df2 = pd.DataFrame([{'qid' : 'q2', 'query' : 'ab', 'docno' : 'd1'}])
            res = cache.transform(df2)
            res1c = cache.transform(df2)
            self.assertEqual(0.4, res1c.iloc[0]["features"][0])
            pd.testing.assert_frame_equal(res, res1c)
        



from typing import Union, List, Tuple, Optional
from pathlib import Path
import pickle
import pandas as pd
import pyterrier as pt
import json
import sqlite3
from collections import defaultdict
from contextlib import closing, contextmanager
import pyterrier_alpha as pta

PICKLE_PROTOCOL = 4

class Sqlite3KeyValueCache(pta.Artifact, pt.Transformer):
    """A cache for storing and retrieving scores for documents, backed by a SQLite3 database.

    This is a *sparse* scorer cache, meaning that space is only allocated for documents that have been scored.
    If a large proportion of the corpus is expected to be scored, a dense cache (e.g., :class:`~pyterrier_caching.Hdf5ScorerCache`)
    may be more appropriate.
    """

    ARTIFACT_TYPE = 'key_value_cache'
    ARTIFACT_FORMAT = 'sqlite3'

    def __init__(
        self,
        path: str,
        transformer: pt.Transformer = None,
        *,
        key: Optional[Union[str, List[str], Tuple[str]]] = None,
        value: Optional[Union[str, List[str], Tuple[str]]] = None,
        verbose: bool = False,
    ):
        """
        Args:
            path: The path to the directory where the cache should be stored.
            transformer: The transformer to apply to get misses from the cache when value isn't found for key.
            key: The name of the column(s) in the input DataFrame that contains the key for looking up value. Can be omitted for caches that have already been created.
            value: The name of the column(s) in the input DataFrame that contains the value value to cache. Can be omitted for caches that have already been created.
            verbose: Whether to print verbose output when scoring documents.

        If a cache does not yet exist at the provided ``path``, a new one is created.

        .. versionadded:: 0.4.0
        """
        super().__init__(path)
        self.transformer = transformer
        self.verbose = verbose
        self.meta = None

        if key is not None:
            if isinstance(key, str):
                key = (key,)
            key = list(key)
        if value is not None:
            if isinstance(value, str):
                value = (value,)
            value = list(value)

        if not (Path(self.path)/'pt_meta.json').exists():
            assert key is not None, "key must be provided when creating a new cache"
            assert value is not None, "value must be provided when creating a new cache"
            with pta.ArtifactBuilder(self) as builder:
                builder.metadata['key'] = key
                builder.metadata['value'] = value
                self.db = sqlite3.connect(builder.path/'db.sqlite3')
                with closing(self.db.cursor()) as cursor:
                    cursor.execute("""
                        CREATE TABLE IF NOT EXISTS cache (
                          key BLOB NOT NULL,
                          value BLOB NOT NULL,
                          PRIMARY KEY (key)
                        )
                    """)
        else:
            self.db = sqlite3.connect(self.path/'db.sqlite3')
        with (Path(self.path)/'pt_meta.json').open('rt') as fin:
            self.meta = json.load(fin)

        if key is None:
            key = self.meta['key']
        assert key == self.meta['key'], f'key={key!r} provided, but index created with key={self.meta["key"]!r}'
        self.key = key

        if value is None:
            value = self.meta['value']
        assert value == self.meta['value'], f'value={value!r} provided, but index created with value={self.meta["value"]!r}'
        self.value = value

    def close(self):
        """ Closes this cache, releasing the sqlite connection that it holds. """
        if self.db is not None:
            self.db.close()
            self.db = None

    @contextmanager
    def _cursor(self):
        assert self.db is not None, "cache is closed"
        with closing(self.db.cursor()) as cursor:
            yield cursor

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        pta.validate.columns(inp, includes=list(self.key))

        to_score_idxs = []
        to_score_map = {}

        inp = inp.reset_index(drop=True)
        keys = [pickle.dumps(tuple(tup), protocol=PICKLE_PROTOCOL) for tup in inp[self.key].itertuples(index=False)]
        values = [[None] * len(inp) for _ in self.value]

        # First pass: load what we can from cache
        key2idxs = defaultdict(list)
        for idx, key in enumerate(keys):
            key2idxs[key].append(idx)
        with self._cursor() as cursor:
            placeholder = ', '.join(['?'] * len(keys))
            cursor.execute(f'SELECT key, value FROM cache WHERE key IN ({placeholder})', keys)
            for key, value in cursor.fetchall():
                value = pickle.loads(value)
                for idx in key2idxs[key]:
                    for v, vs in zip(value, values):
                        vs[idx] = v
                del key2idxs[key]
        for key, idxs in key2idxs.items():
            to_score_idxs.extend(idxs)
            to_score_map[key] = idxs

        # Second pass: score the missing ones and add to cache
        if to_score_idxs:
            if self.transformer is None:
                raise LookupError('values missing from cache, but no transformer provided')
            scored = self.transformer(inp.loc[to_score_idxs])
            keys = scored[self.key]
            keys_enc = [pickle.dumps(tuple(k), protocol=PICKLE_PROTOCOL) for k in keys.itertuples(index=False)]
            value = scored[self.value]
            value_enc = [pickle.dumps(tuple(v), protocol=PICKLE_PROTOCOL) for v in value.itertuples(index=False)]
            with closing(self.db.cursor()) as cursor:
                cursor.executemany('INSERT INTO cache (key, value) VALUES (?, ?)', list(zip(keys_enc, value_enc)))
                self.db.commit()
            for key_enc, value in zip(keys_enc, value.itertuples(index=False)):
                for idx in to_score_map[key_enc]:
                    for v, vs in zip(value, values):
                        vs[idx] = v

        results = inp.assign(**{n: v for n, v in zip(self.value, values)})
        if self.verbose:
            print(f"{self}: {len(inp)-len(to_score_idxs)} hit(s), {len(to_score_idxs)} miss(es)")
        return results

    def __iter__(self):
        with closing(self.db.cursor()) as cursor:
            for key, value in cursor.execute('SELECT key, value FROM cache').fetchall():
                key = pickle.loads(key)
                value = pickle.loads(value)
                res = {}
                res.update(zip(self.key, key))
                res.update(zip(self.value, value))
                yield res

    def __repr__(self):
        return f'Sqlite3KeyValueCache({str(self.path)!r}, {self.transformer!r}, key={self.key!r}, value={self.value!r})'

# Default implementations
KeyValueCache = Sqlite3KeyValueCache



from time import time
import pyterrier as pt
from pyterrier_caching import SparseScorerCache
from pyterrier_t5 import MonoT5ReRanker

# Setup
dataset = pt.get_dataset('irds:msmarco-passage/trec-dl-2019/judged')
bm25 = pt.terrier.Retriever.from_dataset('msmarco_passage', 'terrier_stemmed', wmodel='BM25')
monot5 = pt.text.get_text(dataset, 'text') >> MonoT5ReRanker()
cached_monot5 = SparseScorerCache('monot5.cache', monot5, verbose=True)
inp = bm25(dataset.get_topics())
pipeline = cached_monot5


t0 = time()
pipeline(inp)
t1 = time()
print(f'first invocation: {t1-t0:.3f}s')

t0 = time()
pipeline(inp)
t1 = time()
print(f'second invocation: {t1-t0:.3f}s')
